# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

import cbra.core as cbra
from ..models import AuthorizationRequest
from ..models import AuthorizationRequestParameters
from ..types import AuthorizationRequestIdentifier
from ..types import FatalAuthorizationException
from ..types import IAuthorizationServerStorage


__all__: list[str] = [
    'CookieStoredAuthorizationRequest'
]


async def get(
    storage: IAuthorizationServerStorage = cbra.instance(
        name='_AuthorizationServerStorage'
    ),
    request_id: AuthorizationRequestIdentifier | None = fastapi.Cookie(
        default=None,
        title="Request ID",
        description=(
            "The local authorization request in which the downstream "
            "authentication takes place."
        ),
        alias='oauth2.request'
    ),
    error: str | None = fastapi.Query(
        default=None,
        title="Error code",
        description=(
            "The error code returned by the authorization server if "
            "the user cancelled the request, refused consent, or "
            "failed to authenticate."
        )
    )
):
    if error:
        return None
    params = None
    if request_id is not None:
        params = await storage.get(AuthorizationRequestParameters, request_id)
    if params is None:
        raise FatalAuthorizationException(
            f"The authorization request (id: {request_id}) does not "
            "exist, is expired, or otherwise not processable."
        )
    authnrequest = AuthorizationRequest.parse_obj(params)
    return authnrequest


CookieStoredAuthorizationRequest: AuthorizationRequest = fastapi.Depends(get)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from cbra.types import PolicyPrincipal


class Binding(pydantic.BaseModel):
    role: str = pydantic.Field(
        default=...,
        title="Role",
        description=(
            "Role that is assigned to the list of `members`, or principals. "
            "For example, `roles/viewer`, `roles/editor`, or `roles/owner`."
        )
    )

    members: list[PolicyPrincipal] = pydantic.Field(
        default=[],
        title="Members",
        description=(
            "Specifies the principals requesting access for a resource. "
            "The `members` array can have the following values:\n\n"
            "- **user:{email}**: an email address that is associated to "
            "a specific account."
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

from canonical import EmailAddress

from cbra.types import SessionClaims


class IAuthorizationRequest(Protocol):

    @property
    def auth_time(self) -> int | None:
        ...

    @property
    def email(self) -> EmailAddress | None:
        ...

    @property
    def email_verified(self) -> bool:
        ...

    @property
    def session(self) -> SessionClaims:
        ...

    def has_owner(self) -> bool: ...


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import importlib

from .container import Container
from .dependency import Dependency
from .injected import Injected


class Setting(Injected):
    __module__: str = 'cbra.core.ioc'

    def add_to_container(self, container: Container) -> None:
        if container.has(self.name):
            # TODO: This basically cover the user case where the implementer
            # override the dependency from somewhere in the code. This is
            # currently only used in the examples.
            self.ref = container.require(self.name)
            return
        try:
            config = importlib.import_module('cbra.core.conf')
            self.ref = Dependency(
                name=self.name,
                qualname=f'cbra.core.conf.settings.{self.name}',
                symbol=getattr(config.settings, self.name)
            )
        except (AttributeError, ImportError):
            if self.missing == NotImplemented:
                raise
            self.ref = Dependency(
                name=self.name,
                qualname=f'cbra.core.conf.settings.{self.name}',
                symbol=self.missing
            )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import copy
import os
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import TypeVar
from typing import Union

import fastapi
import pydantic
from canonical import EmailAddress
from headless.ext.oauth2.models import OIDCToken
from headless.ext.oauth2.types import ResponseMode
from headless.ext.oauth2.types import ResponseType

from cbra.types import IDependant
from cbra.types import Request
from cbra.types import SessionClaims
from cbra.core.iam.models import Subject
from cbra.core.ioc import override
from cbra.core.conf import settings
from cbra.core.sessions import RequestSession
from .authorizationrequestobject import AuthorizationRequestObject
from .authorizationrequestparameters import AuthorizationRequestParameters
from .authorizationrequestreference import AuthorizationRequestReference
from .authorizationrequestclient import AuthorizationRequestClient
from .resourceowner import ResourceOwner
from ..types import AccessType
from ..types import ClientInfo
from ..types import FatalAuthorizationException
from ..types import FrontendException
from ..types import IAuthorizationServerStorage
from ..types import InvalidResponseTypeRequested
from ..types import ITokenBuilder
from ..types import RedirectURI
from ..types import RequestedScope
from ..types import ResourceOwnerIdentifier
from ..types import MissingResponseType


T = TypeVar('T', bound='BaseAuthorizationRequest')


class BaseAuthorizationRequest(IDependant, pydantic.BaseModel):

    @staticmethod
    def get_issuer(request: Request) -> str:
        return settings.OAUTH2_ISSUER or\
            f'{request.url.scheme}://{request.url.netloc}'

    @staticmethod
    def get_remote_host(request: Request) -> str:
        assert request.client is not None
        return str(request.client.host)

    @classmethod
    def fromrequest(
        cls: type[T],
        client: AuthorizationRequestClient = AuthorizationRequestClient.depends(),
        iss: str = fastapi.Depends(get_issuer),
        remote_host: str = fastapi.Depends(get_remote_host),
        access_type: AccessType = fastapi.Query(
            default=AccessType.online,
            title="Access type",
            description=(
                "Indicates whether your application can refresh access tokens "
                "when the user is not present at the browser. Valid parameter "
                "values are `online`, which is the default value, and `offline`."
            )
        ),
        client_id: str | None = fastapi.Query(
            default=None,
            title="Client ID",
            description="Identifies the client that is requesting authorization."
        ),
        response_type: ResponseType | str | None = fastapi.Query(
            default=None,
            title="Response type",
            description=(
                "Informs the authorization server of the desired response type. "
                "This parameter is required."
            ),
            example="code",
        ),
        redirect_uri: RedirectURI | None = fastapi.Query(
            default=None,
            title="Redirect URI",
            description=(
                "The URL to redirect the client to after completing the "
                "flow. Must be an absolute URI that is served over https, if "
                "not redirecting to `localhost`.\n\n"
                "If `redirect_uri` is omitted, the default redirect URI for "
                "the client specified by `client_id` is used. For clients that "
                "do not have a redirect URI specified, this produces an error "
                "state."
            )
        ),
        scope: str | None = fastapi.Query(
            default=None,
            title="Scope",
            description=(
                "A space-delimited list specifying the requested access scope."
            ),
            max_length=512,
            example="hello.world"
        ),
        state: str | None = fastapi.Query(
            default=None,
            title="State",
            description=(
                "An opaque value used by the client to maintain state between "
                "the request and callback. The authorization server includes "
                "this value when redirecting the user-agent back to the client."
            ),
            max_length=64,
            example=bytes.hex(os.urandom(64))
        ),
        response_mode: ResponseMode | None = fastapi.Query(
            default=None,
            title="Response mode",
            description=(
                "Informs the authorization server of the mechanism to be used "
                "for returning authorization response parameters."
            ),
            example="query"
        ),
        resource: list[str] = fastapi.Query(
            default=[],
            title="Resource",
            description=(
                "Indicates the target service or resource to which access is "
                "being requested. Its value **must** be an absolute URI, as "
                "specified by Section 4.3 of RFC 3986. The URI **must not** "
                "include a fragment component. It **should not** include a query "
                "component, but it is recognized that there are cases that make a "
                "query component a useful and necessary part of the resource "
                "parameter, such as when one or more query parameters are used "
                "to scope requests to an application. The resource parameter "
                "URI value is an identifier representing the identity of the "
                "resource, which **may** be a locator that corresponds to a "
                "network-addressable location where the target resource is "
                "hosted. Multiple resource parameters **may** be used to "
                "indicate that the requested token is intended to be used at "
                "multiple resources."
            )
        ),
        request: str | None = fastapi.Query(
            default=None,
            title="Request",
            description=(
                "A JSON Web Token (JWT) whose JWT Claims Set holds the "
                "JSON-encoded OAuth 2.0 authorization request parameters. "
                "Must not be used in combination with the `request_uri` "
                "parameter, and all other parameters except `client_id` "
                "must be absent.\n\n"
                "Confidential and credentialed clients must first sign "
                "the claims using their private key, and then encrypt the "
                "result with the public keys that are provided by the "
                "authorization server through the `jwks_uri` specified "
                "in its metadata."
            )
        ),
        request_uri: str | None = fastapi.Query(
            default=None,
            title="Request URI",
            description=(
                "References a Pushed Authorization Request (PAR) or a remote "
                "object containing the authorization request.\n\n"
                "If the authorization request was pushed to this authorization "
                "server, then the format of the `request_uri` parameter is "
                "`urn:ietf:params:oauth:request_uri:<reference-value>`. "
                "Otherwise, it is an URI using the `https` scheme. If the "
                "`request_uri` parameter is a remote object, then the external "
                "domain must have been priorly whitelisted by the client."
            )
        ),

        # OpenID parameters
        nonce: str | None = fastapi.Query(
            default=None,
            title="Nonce",
            description=(
                "String value used to associate a Client session with an ID Token, "
                "and to mitigate replay attacks. The value is passed through "
                "unmodified from the Authentication Request to the ID Token. "
                "Sufficient entropy MUST be present in the nonce values used "
                "to prevent attackers from guessing values."
            )
        )
    ) -> T:
        raise NotImplementedError

    @classmethod
    def __inject__(cls: type[T]) -> Callable[..., Awaitable[T] | T]:
        return cls.fromrequest


class AuthorizationRequest(BaseAuthorizationRequest):
    __root__: Union[
        AuthorizationRequestReference,
        AuthorizationRequestObject,

        # This needs to be the last value here because it is responsible
        # for raising the abortable.
        AuthorizationRequestParameters,
    ]

    @classmethod
    @override(BaseAuthorizationRequest.fromrequest) # type: ignore
    def fromrequest(cls: type[T], **kwargs: Any) -> T: # type: ignore
        iss: str = kwargs.pop('iss')
        client: AuthorizationRequestClient = kwargs.pop('client')
        try:
            return cls.parse_obj({
                **copy.deepcopy(kwargs),
                'scope': kwargs.get('scope') or set(),
                'client': client,
                'client_info': client.client_info
            })
        except pydantic.ValidationError as e:
            state: str | None = kwargs.get('state')
            redirect_uri = client.get_redirect_uri(kwargs.get('redirect_uri'))

            assert isinstance(redirect_uri, RedirectURI)
            for error in e.errors():
                loc = (error.get('loc') or (None, None))[-1]
                typ = error.get('type') or 'error'
                if loc == 'response_type' and typ == 'type_error.none.not_allowed':
                    raise MissingResponseType(redirect_uri, iss, state)
                if loc == 'response_type' and typ == 'type_error.enum':
                    raise InvalidResponseTypeRequested(redirect_uri, iss, state)
            raise Exception(e)

    @property
    def auth_time(self) -> int:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        assert self.__root__.auth_time is not None
        return self.__root__.auth_time

    @property
    def client_id(self) -> str:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return self.__root__.client_id

    @property
    def client_info(self) -> ClientInfo:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return self.__root__.client_info

    @property
    def code(self) -> str | None:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return str(self.__root__.code) if self.__root__.code else None

    @property
    def consent(self) -> list[RequestedScope]:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return self.__root__.consent

    @property
    def email(self) -> EmailAddress | None:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return self.__root__.email

    def set_email(self, value: str) -> None:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        self.__root__.email =  EmailAddress(value)

    @property
    def email_verified(self) -> bool:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return self.__root__.email_verified or False

    @property
    def id(self) -> str:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        assert self.__root__.id is not None
        return self.__root__.id

    @property
    def id_token(self) -> dict[str, Any]:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return self.__root__.id_token

    @property
    def nonce(self) -> str | None:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return self.__root__.nonce

    @property
    def owner(self) -> ResourceOwnerIdentifier:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        assert self.__root__.owner is not None
        return ResourceOwnerIdentifier(
            client_id=self.__root__.client_id,
            sub=self.__root__.owner
        )

    @property
    def redirect_uri(self) -> str:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return self.__root__.redirect_uri

    @property
    def request_uri(self) -> str:
        return f'urn:ietf:params:oauth:request_uri:{self.id}'

    @property
    def resources(self) -> set[str]:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return set(self.__root__.resource)

    @property
    def scope(self) -> list[RequestedScope]:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return self.__root__.scope

    @property
    def session_id(self) -> str | None:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return self.__root__.session_id

    @property
    def session(self) -> SessionClaims:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        assert self.__root__.session is not None
        return self.__root__.session

    def authenticate(self, oidc: OIDCToken) -> None:
        """Authenticates the request using the external OIDC token."""
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        if self.is_authenticated():
            raise ValueError("Request is already authenticated")
        assert not self.__root__.downstream
        self.__root__.email = oidc.email
        self.__root__.downstream = oidc

    def as_response(self, client: Any, iss: str) -> fastapi.Response:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return self.__root__.as_response(client, iss)

    def add_to_session(self, session: RequestSession):
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        self.__root__.session_id = session.id
        self.__root__.session = session.claims

    def clone(self):
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return self.parse_obj(self.__root__.get_parameters())

    def get_authorize_url(self, request: fastapi.Request) -> str:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return (
            f"{request.url_for('oauth2.authorize')}?"
            f"request_uri={self.request_uri}&"
            f"client_id={self.__root__.client_id}"
        )
    
    def has_id_token(self) -> bool:
        return 'openid' in {x.name for x in self.scope}

    def has_owner(self) -> bool:
        """Return a boolean indicating if the request has an owner."""
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return self.__root__.owner is not None

    def has_refresh_token(self) -> bool:
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return self.__root__.access_type == AccessType.offline

    def is_authenticated(self) -> bool:
        """Return a boolean indicating if the authorization request was
        authenticated by a downstream identity provider.
        """
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return self.__root__.downstream is not None
    
    def is_owned_by(self, uid: int) -> bool:
        """Return a boolean indicating if the request is owned by the
        user.
        """
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return self.__root__.owner == uid

    def must_consent(self) -> bool:
        """Return a boolean indicating if the resource owner must consent to
        the scope/claims requested.
        """
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        return bool(self.__root__.consent)

    def needs(self, subject: Subject) -> set[str]:
        """Return the set of claims that the Resource Owner must provide
        in order to continue with the authorization request.
        """
        ignore: set[str] = {"email"}
        required, _ = self.wants()
        return {
            x for x in required
            if not subject.has_claim(x)
            and x not in ignore
        }

    def wants(self) -> tuple[set[str], set[str]]:
        """Return a tuple containing the set of required claims, and the
        set of optional claims.
        """
        required: set[str] = set()
        optional: set[str] = set()
        for requested in self.scope:
            scope_required, scope_optional = requested.wants()
            required |= scope_required
            optional |= scope_optional
        return required, optional - required

    async def load(
        self,
        client: Any,
        storage: Any,
        session_id: str
    ) -> None:
        """Resolve pushed requests or request references so that :attr:`__root__`
        is always an intance of :class:`AuthorizationRequestParameters`.
        """
        self.__root__ = await self.__root__.load(client, storage, session_id)
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        assert self.__root__.id is not None
        self.__root__.session_id = session_id
        await self.persist(storage)

    async def persist(self, storage: IAuthorizationServerStorage) -> None:
        await storage.persist(self.__root__)

    async def verify(
        self,
        builder: ITokenBuilder,
        request: Request,
        session: SessionClaims,
        client: AuthorizationRequestClient,
        owner: ResourceOwner,
        subject: Subject
    ) -> None:
        """Verifies that all parameters in the request are valid, supported
        and allowed by the client, and allowed by the subject.
        """
        sub = owner.sub
        assert session.auth_time is not None
        assert session.email is not None
        assert isinstance(self.__root__, AuthorizationRequestParameters)
        email = self.email or owner.email or session.email
        email_verified = session.email_verified or self.__root__.email_verified
        if not client.allows_email(email):
            raise FrontendException(
                '/select-email',
                next=self.get_authorize_url(request),
            )
        if self.resources and not client.allows_resources(set(self.resources)):
            raise FatalAuthorizationException(
                "The client does not allow authorization for the "
                "requested resource(s)."
            )
        if not client.can_redirect(self.__root__.redirect_uri):
            raise FatalAuthorizationException(
                "The client does not allow redirection to the given "
                "redirect_uri.",
                status_code=403
            )
        if not client.can_use(self.__root__.scope):
            raise FatalAuthorizationException(
                "The client does not allow use of the given scope.",
                status_code=403
            )

        if self.has_id_token() and self.nonce is None: # is openid
            raise FatalAuthorizationException(
                "The 'nonce' parameter is required for OpenID authorization "
                "requests."
            )

        self.__root__.assign(sub)
        self.__root__.session = session
        self.__root__.auth_time = session.auth_time
        self.__root__.email = email
        self.__root__.email_verified = email_verified
        self.__root__.consent = owner.consent_required(self.scope)

        if self.has_id_token():
            # Create an unsigned ID Token that we can display to the end-user.
            assert self.nonce is not None
            assert owner.ppid.value is not None
            assert self.email is not None
            token = builder.id_token(
                subject=subject,
                ppid=owner.ppid.value,
                nonce=self.nonce,
                scope=self.scope,
                access_token=None,
                auth_time=session.auth_time,
                authorization_code=None,
                request=self,
                owner=owner
            )
            self.__root__.id_token = token.claims


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from typing import Any
from typing import Protocol
from typing import TypeVar

from canonical import DomainName
from canonical import EmailAddress
from headless.ext.oauth2.models import OIDCToken

from cbra.types import ISessionManager
from .principaltype import PrincipalType


T = TypeVar('T', bound='Subject')


class Subject(Protocol):
    uid: int | None
    seen: datetime

    def activate(self) -> None: ...
    def add_principal(self, issuer: str, value: PrincipalType, asserted: datetime, trust: bool = False) -> None: ...
    def add_to_session(self, session: ISessionManager[Any]) -> None: ...
    def can_destroy(self) -> bool: ...
    def can_select_email(self) -> bool: ...
    def get_email(self) -> EmailAddress: ...
    def get_principals(self) -> list[Any]: ...
    def has_claim(self, claim: str) -> bool: ...
    def has_principal(self, principal: PrincipalType) -> bool: ...
    def is_active(self) -> bool: ...
    def merge(self: T, other: T) -> None: ...
    def needs_fallback_email(self, allow: set[DomainName]) -> bool: ...
    def update_oidc(self, oidc: OIDCToken) -> None: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi
from headless.ext import oauth2
from headless.types import IClient

from cbra.core.conf import settings
from cbra.types import NotAuthorized
from cbra.types import NotFound
from ..types import IFrontendStorage
from ..types import ManagedGrantIdentifier
from .applicationclient import ApplicationClient
from .clientstorage import ClientStorage


async def get(
    client: oauth2.Client = ApplicationClient,
    storage: IFrontendStorage = ClientStorage,
    grant_id: ManagedGrantIdentifier | None = fastapi.Cookie(
        default=None,
        title="Grant ID",
        alias='bff.grant',
        description=(
            "Identifies the grant that is usesd to exchange tokens with "
            "the authorization server."
        )
    ),
    resource: str = fastapi.Path(
        default=...
    ),
    path: str = fastapi.Path(
        default=...
    )
):
    path = path or '/'
    spec = settings.OAUTH2_RESOURCE_SERVERS.get(resource)
    if spec is None:
        raise NotFound
    resource = spec['resource']
    if grant_id is None:
        raise NotAuthorized(
            headers={'WWW-Authenticate': 'authorization_code'}
        )
    grant = await storage.get(grant_id)
    if grant is None:
        raise NotAuthorized(
            headers={'WWW-Authenticate': 'authorization_code'}
        )
    async with await grant.get_resource_client(storage, client, resource, spec.get('scope'))\
    as resource_client:
        yield resource_client


RequestedResourceServerClient: IClient[Any, Any] = fastapi.Depends(get)




# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol
from typing import TypeVar

from canonical import EmailAddress

from cbra.types import IEmailSender


T = TypeVar('T', bound='IEmailChallenge')


class IEmailChallenge(Protocol):
    @classmethod
    def new(cls: type[T], email: EmailAddress) -> T: ...
    def age(self) -> int: ...
    def get_challenge_id(self) -> str: ...
    def is_blocked(self) -> bool: ...
    def verify(self, code: str) -> bool: ...

    async def send(
        self,
        service: IEmailSender,
        sender: str,
        subject: str,
        template: str = 'email-verification.html.j2',
        context: dict[str, Any] | None = None
    ) -> None:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from cbra.types import Abortable


class FatalAuthorizationException(Abortable):
    __module__: str = 'cbra.ext.oauth2.types'
    status_code: int = 400

    def __init__(self, message: str, status_code: int = 400):
        self.message = message
        self.status_code = status_code

    async def as_response(self) -> fastapi.Response:
        return fastapi.responses.PlainTextResponse(
            status_code=self.status_code,
            content=self.message
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol

import aorta


class IOriginStorage(Protocol):
    __module__: str = 'cbra.ext.origin.types'

    async def latest(self) -> aorta.types.Envelope[Any]:
        ...

    async def persist(
        self,
        message: aorta.types.Envelope[Any]
    ) -> None:
        ...

    async def seen(
        self,
        message: aorta.types.Envelope[Any]
    ) -> bool:
        ...


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
import uvicorn

import cbra.core as cbra
from cbra.core.iam import AuthorizationContextFactory
from cbra.types import Principal


app: cbra.Application = cbra.Application()


@app.get('/')
async def f(
    request: fastapi.Request,
    factory: AuthorizationContextFactory = fastapi.Depends(),
    principal: Principal = fastapi.Depends(Principal.fromrequest)
):
    ctx = await factory.authenticate(request, principal)
    return {
        'remote_host': ctx.remote_host,
        'is_authenticated': ctx.is_authenticated()
    }


if __name__ == '__main__':
    uvicorn.run('__main__:app', reload=True) # type: ignore



# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
from typing import Any

import uvicorn

os.environ['PYTHON_SETTINGS_MODULE'] = __name__
import cbra.core as cbra


DEPENDENCIES: list[dict[str, Any]] = [
    {
        'name': 'RandomGenerator',
        'qualname': 'os.urandom'
    },
    {
        'name': 'SymbolRequirement',
        'qualname': 'cbra.core.ioc.test.SymbolRequirement'
    },
    {
        'name': 'InstanceRequirement',
        'qualname': 'cbra.core.ioc.test.InstanceRequirement'
    },
    {
        'name': 'NestedRequirement',
        'qualname': 'cbra.core.ioc.test.NestedRequirement'
    },
    {
        'name': 'SingletonRequirement',
        'qualname': 'cbra.core.ioc.test.SingletonRequirement',
        'singleton': True
    },
]

app = cbra.Application(docs_url='/ui')


@app.get('/')
def f(
    injected: Any = cbra.inject('RandomGenerator'),
    symbol: Any = cbra.inject('SymbolRequirement'),
    instance: Any = cbra.instance('InstanceRequirement'),
    nested: Any = cbra.instance('NestedRequirement'),
    singleton: Any = cbra.instance('SingletonRequirement')
):
    print(symbol, instance, nested, singleton)
    return bytes.hex(injected(16))

if __name__ == '__main__':
    uvicorn.run('__main__:app', reload=True) # type: ignore



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from ..types import AuthorizationRequestIdentifier
from ..types import FatalAuthorizationException
from .authorizationrequestparameters import AuthorizationRequestParameters


class AuthorizationRequestReference(pydantic.BaseModel):
    client_id: str
    request_uri: str
    remote_host: str

    @property
    def id(self) -> str:
        return AuthorizationRequestIdentifier(str.split(self.request_uri, ':')[-1])

    async def load(
        self,
        client: Any,
        storage: Any,
        session_id: str
    ) -> AuthorizationRequestParameters:
        params = await storage.get(AuthorizationRequestParameters, self.id)
        if params is None:
            raise FatalAuthorizationException("The request does not exist.")
        return params


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic


class ResourceListMetadata(pydantic.BaseModel):
    #pagination_token: str = pydantic.Field(
    #    default="",
    #    alias='paginationToken',
    #    title='Pagination token',
    #    description=(
    #        'This field represents the pagination token to '
    #        'retrieve the next page of results. If the value '
    #        'is "", it means no further results for the request.'
    #    )
    #)

    next_url: str | None = pydantic.Field(
        default=None,
        alias='nextUrl',
        title="Next URL",
        description=(
            "If the server returned a partial set of the resources "
            "matching the search predicate, the URL at which the next "
            "part of the set can be retrieved. If `nextUrl` is `null`, "
            "them there are no remaining items."
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from datetime import datetime
from datetime import timezone
from typing import Any

import pydantic
from headless.types import IClient
from headless.ext import oauth2
from ..client import AccessTokenFactory
from ..types import BearerToken
from ..types import BearerTokenCredential
from ..types import ResourceAccessTokenIdentifier
from ..types import ResourceServerAccessToken
from ..types import IFrontendStorage


class ManagedGrant(pydantic.BaseModel):
    """A grant received from an external identity
    provider.
    """
    client_id: str
    id: str = pydantic.Field(
        default_factory=lambda: secrets.token_urlsafe(96)
    )
    iss: str
    received: datetime = pydantic.Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )
    refresh_token: str
    scope: set[str] = set()
    sub: str | None = None
    version: int = 0

    def get_access_token_id(
        self,
        resource: str
    ) -> ResourceAccessTokenIdentifier:
        """Return a :class:`~cbra.ext.oauth2.types.ResourceAccessTokenIdentifier`
        instance identifying an access token from this grant for the specified
        resource.
        """
        return ResourceAccessTokenIdentifier(grant_id=self.id, resource=resource)

    async def refresh(
        self,
        storage: IFrontendStorage,
        client: oauth2.Client,
        resource: str | None = None,
        scope: set[str] | None = None
    ) -> BearerToken:
        """Refreshes the refresh token using the given client."""
        response = await client.refresh_token(
            token=self.refresh_token,
            resource=resource,
            scope=scope
        )
        if response.refresh_token is not None:
            self.refresh_token = response.refresh_token
        if resource is not None:
            assert client.server.metadata is not None
            assert client.server.metadata.issuer is not None
            at = ResourceServerAccessToken.parse_response(
                grant_id=self.id,
                issuer=client.server.metadata.issuer,
                response=response,
                resource=resource,
            )
            await storage.persist(at)
        await storage.persist(self)
        return BearerToken.parse_token_response(response)

    async def get_resource_client(
        self,
        storage: IFrontendStorage,
        client: oauth2.Client,
        resource: str,
        scope: set[str] | None = None
    ) -> IClient[Any, Any]:
        token = await storage.get(self.get_access_token_id(resource))
        if token is None:
            token = await self.refresh(storage, client, resource, scope=scope)
        return oauth2.ResourceServer(
            base_url=resource,
            credential=BearerTokenCredential(
                factory=AccessTokenFactory(
                    storage=storage,
                    client=client,
                    resource=resource,
                    grant=self
                ),
                access_token=token
            )
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
import re

import fastapi
import pydantic
from headless.ext.oauth2.types import GrantType

import cbra.core as cbra
from cbra.core.iam.types import Subject
from ..authorizationserverstorage import AuthorizationServerStorage
from ..models import Client
from ..models import Grant
from ..models import ResourceOwner
from ..types import AuthorizationCode
from ..types import FatalClientException
from ..types import IAuthorizationRequest
from ..types import InvalidRequest
from ..types import RefreshTokenIdentifier
from .requestingclient import RequestingClient


logger: logging.Logger = logging.getLogger('cbra.endpoint')


async def get(
    storage: AuthorizationServerStorage = cbra.instance('_AuthorizationServerStorage'),
    client: Client = RequestingClient,
    grant_type: str = fastapi.Form(
        default=...,
        title="Grant type"
    ),
    code: AuthorizationCode | None = fastapi.Form(
        default=None,
        title="Code",
        description=(
            "The authorization code received from the authorization server. "
            "This parameter is **required** when `grant_type` is "
            "`authorization_code`, otherwise it is ignored."
        )
    ),
    redirect_uri: str | None = fastapi.Form(
        default=None,
        title="Redirect URI",
        description=(
            "This parameter is **required**, if the `redirect_uri` parameter "
            "was included in the authorization request, in which case their "
            "values **must"" be identical.  If no `redirect_uri` was included "
            "in the authorization request, this parameter is **optional**."
        )
    ),
    refresh_token: RefreshTokenIdentifier | None = fastapi.Form(
        default=None,
        title="Refresh token",
        description=(
            "The refresh token issued to the client. This parameter is "
            "**required** if `grant_type` is `refresh_token`."
        )
    ),
    resource: str | None = fastapi.Form(
        default=None,
        title="Resource",
        description=(
            "Indicates the target service or protected resource where the "
            "client intends to use the requested access token."
        )
    ),
    scope: str | None = fastapi.Form(
        default=None,
        title="Scope",
        description=(
            "The scope of the access request. "
            "The requested scope MUST NOT include any scope not "
            "originally granted by the resource owner, and if "
            "omitted is treated as equal to the scope originally "
            "granted by the resource owner."
        )
    )
):
    try:
        grant_type = GrantType(grant_type)
    except ValueError:
        raise FatalClientException(
            code='unsupported_grant_type',
            message=(
                "The authorization server does not support the given grant:"
                f"{grant_type[:64]}"
            )
        )
    if not client.can_grant(grant_type):
        raise FatalClientException(
            code='unauthorized_client',
            message=(
                "The client does not allow or is not allowed to use the "
                f"given grant: {grant_type[:64]}"
            )
        )
    
    authnrequest: IAuthorizationRequest | None = None
    owner: ResourceOwner | None = None
    subject: Subject | None = None
    if grant_type == GrantType.authorization_code:
        if code is None:
            raise InvalidRequest("The authorization code is not accepted by the server.")
        authnrequest = await storage.fetch(code)
    if authnrequest is not None:
        assert authnrequest.has_owner()
        owner = await storage.get(ResourceOwner, authnrequest.owner)
        subject = await storage.get(Subject, authnrequest.owner.sub)
    if owner is None:
        logger.critical("Resource owner not found (client: %s)", client.client_id)
    if subject is None:
        logger.critical("Subject not found (client: %s)", client.client_id)
    try:
        return Grant.parse_obj({
            'grant_type': grant_type,
            'code': code,
            'client': client,
            'owner': owner,
            'redirect_uri': redirect_uri,
            'refresh_token': refresh_token,
            'resource': resource,
            'request': authnrequest,
            'scope': set(filter(bool, re.split(r'\s+', scope or ''))),
            'subject': subject
        })
    except pydantic.ValidationError as e:
        logger.debug("Caught fatal ValidationError: %s", repr(e))
        raise InvalidRequest(
            f"The parameters for grant {grant_type[:64]} were invalid, "
            f"incomplete, duplicated or otherwise malformed."
        )


RequestedGrant: Grant = fastapi.Depends(get)


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import base64
import secrets
from typing import Any
from typing import TypeVar

import pydantic

from .ihashable import IHashable
from .sessionclaims import SessionClaims


T = TypeVar('T', bound='SessionModel')


class SessionModel(pydantic.BaseModel, IHashable):
    id: str
    iat: int
    claims: SessionClaims | None = None
    csrf: str | None = pydantic.Field(
        default_factory=lambda: secrets.token_urlsafe(48)
    )
    hmac: str | None = None

    @classmethod
    def parse_cookie(cls: type[T], value: str | None) -> T | None:
        if value is None: return None
        try:
            serialized = base64.urlsafe_b64decode(str.encode(value, 'ascii'))
            return cls.parse_raw(serialized)
        except Exception:
            return None

    def digest(self) -> bytes:
        return self._hash(self.dict(exclude={'hmac'}, exclude_none=True))

    def get(self, key: str) -> Any:
        if not self.claims:
            return None
        if key not in self.claims.__fields__:
            raise AttributeError(f'No such claim: {key}')
        return getattr(self.claims, key)


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any


class Persister:

    async def persist(
        self,
        resource: Any,
        create: bool = False
    ) -> Any:
        raise NotImplementedError(
            f'Implement {type(self).__name__}.persist() '
            f'to persist {type(resource).__name__} resources.'
        )


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Coroutine

import pytest

from cbra.ext import webhooks


__all__: list[str] = [
    'test_handler_exception',
    'test_send_message',
    'test_unknown_format'
]


@pytest.mark.asyncio
async def test_send_message(
    request_factory: Callable[..., Coroutine[Any, Any, webhooks.WebhookResponse]],
    sign: bool
):
    result = await request_factory(
        topic='hello',
        msg={'foo': 1},
        url='/',
        sign=sign
    )
    assert result.success, result.reason
    assert result.accepted


@pytest.mark.asyncio
async def test_handler_exception(
    request_factory: Callable[..., Coroutine[Any, Any, webhooks.WebhookResponse]],
    sign: bool
):
    result = await request_factory(
        topic='exception',
        msg={'foo': 1},
        url='/',
        sign=sign
    )
    assert not result.success
    assert not result.accepted
    assert result.code == 'FATAL_EXCEPTION'


@pytest.mark.asyncio
async def test_unknown_format(
    request_factory: Callable[..., Coroutine[Any, Any, webhooks.WebhookResponse]],
    sign: bool
):
    result = await request_factory(
        topic='hello',
        msg={'_': 1},
        url='/',
        sign=sign
    )
    assert not result.success
    assert not result.accepted
    assert result.code == 'UNKNOWN_MESSAGE_FORMAT'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from typing import TypeVar

import pydantic
from headless.ext.oauth2.models import TokenResponse


T = TypeVar('T', bound='BearerToken')


class BearerToken(pydantic.BaseModel):
    access_token: str
    expires: datetime

    @classmethod
    def parse_token_response(
        cls: type[T],
        response: TokenResponse,
        leeway: int = 30
    ) -> T:
        return cls.parse_obj({
            'access_token': str(response.access_token),
            'expires': datetime.now(timezone.utc)\
                + timedelta(seconds=response.expires_in - leeway)
        })

    def is_expired(self) -> bool:
        return datetime.now(timezone.utc) >= self.expires
    
    def __str__(self) -> str:
        return self.access_token


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generic
from typing import Sequence
from typing import TypeVar

import pydantic
from googleapiclient import discovery


T = TypeVar('T', bound='SpreadsheetModel')


class SpreadsheetModel(pydantic.BaseModel):

    @classmethod
    def parse_row(
        cls: type[T],
        row: Sequence[Any]
    ) -> T:
        values: dict[str, Any] = {}
        for i, k in enumerate(cls.__fields__):
            try:
                values[k] = row[i]
            except IndexError:
                values[k] = None
        return cls.parse_obj(values)

    def as_row(self) -> dict[str, Any]:
        values = tuple(getattr(self, k) for k in self.__fields__.keys())
        return {'values': [
            {'userEnteredValue': {'stringValue': str(x or '')}}
            for x in values
        ]}


class Spreadsheet(Generic[T]):
    __module__: str = 'ea.ui'
    id: str
    model: type[T]
    service: Any = discovery.build('sheets', 'v4') # type: ignore
    sheet_id: int

    def __init__(
        self,
        model: type[T],
        id: str,
        sheet_id: int,
        range: str = 'A:Z'
    ):
        self.client = self.service.spreadsheets()
        self.id = id
        self.model = model
        self.range = range
        self.sheet_id = sheet_id
        self.values = []

    def load(self) -> None:
        """Load the current values."""
        request = self.client.values().get(
            spreadsheetId=self.id,
            range=self.range
        )
        response = request.execute()
        self.values = [
            self.model.parse_row(x)
            if any(x) else None
            for x in ((response.get('values') or [])[1:])
        ]

    def append(self, row: T) -> None:
        return self.extend([row])

    def extend(self, rows: list[T]) -> None:
        request = self.client.batchUpdate(
            spreadsheetId=self.id,
            body={
                'requests': [{
                    'appendCells': {
                        'sheetId': self.sheet_id,
                        'fields': '*',
                        'rows': [x.as_row() for x in rows]
                    }
                }]
            }
        )
        request.execute()
        self.write()

    def clear(self) -> None:
        """Clears all rows in the spreadsheet."""
        self.client.batchUpdate(
            spreadsheetId=self.id,
            body={
                'requests': [
                    {
                        'updateSheetProperties': {
                            'properties': {
                                'sheetId': self.sheet_id,
                                'gridProperties': {
                                    'frozenRowCount': 0
                                }
                            },
                            'fields': 'gridProperties.frozenRowCount'
                        }
                    }
                ]
            }
        ).execute()
        self.client.batchUpdate(
            spreadsheetId=self.id,
            body={
                'requests': {
                    'deleteDimension': {
                        'range': {
                            'sheetId': self.sheet_id,
                            'dimension': "ROWS",
                            'startIndex': 1
                        }
                    }
                }
            }
        ).execute()

    def write(self) -> None:
        self.client.batchUpdate(
            spreadsheetId=self.id,
            body={
                'requests': [
                    {
                        'updateSheetProperties': {
                            'properties': {
                                'sheetId': self.sheet_id,
                                'gridProperties': {
                                    'frozenRowCount': 1
                                }
                            },
                            'fields': 'gridProperties.frozenRowCount'
                        }
                    },
                    {
                        'repeatCell': {
                            'fields': "userEnteredFormat(textFormat)",
                            'range': {
                                'startRowIndex': 1,
                                'sheetId': self.sheet_id
                            },
                            'cell': {
                                'userEnteredFormat': {
                                    'textFormat': {
                                        'fontSize': 10,
                                        'foregroundColor': {
                                            'red': 0,
                                            'green': 0,
                                            'blue': 0
                                        }
                                    }
                                }
                            }
                        }
                    }
                ]
            }
        ).execute()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi
from headless.types import IClient

from .params import RequestedResourceServerClient
from .tokenhandlerendpoint import TokenHandlerEndpoint


class FrontendProxyEndpoint(TokenHandlerEndpoint):
    __module__: str = 'cbra.ext.oauth2'
    name: str = 'bff.proxy'
    path: str = '/oauth/v2/resources/{resource}{path:path}'
    resource: IClient[Any, Any] = RequestedResourceServerClient
    resource_path: str = fastapi.Path(default=..., alias='path')
    summary: str = 'Frontend Proxy Endpoint'

    async def head(self) -> fastapi.Response:
        return await self.handle('HEAD')

    async def get(self) -> fastapi.Response:
        return await self.handle('GET')

    async def post(self) -> fastapi.Response:
        return await self.handle('POST')

    async def put(self) -> fastapi.Response:
        return await self.handle('PUT')

    async def patch(self) -> fastapi.Response:
        return await self.handle('PATCH')

    async def delete(self) -> fastapi.Response:
        return await self.handle('DELETE')
    
    async def handle(self, method: str) -> fastapi.Response:
        response = await self.resource.request(method=method, url=self.resource_path)
        return fastapi.Response(
            content=response.content,
            headers=dict(response.headers),
            status_code=response.status_code
        )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .icredential import ICredential


class JSONWebToken(ICredential):
    """An unparsed JSON Web Token."""
    __module__: str = 'cbra.types'
    token: str

    def __init__(self, token: str):
        self.token = token

    def __str__(self) -> str:
        return self.token


#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
from headless.ext.oauth2.models import TokenResponse

from cbra.types import Request
from ..types import OIDCProvider
from ..types import FatalAuthorizationException
from ..types import IExternalAuthorizationState
from .localopenidprovider import LocalOpenIdProvider
from .localauthorizationrequeststate import LocalAuthorizationRequestState


__all__: list[str] = [
    'DownstreamTokenResponse'
]

async def get(
    request: Request,
    provider: OIDCProvider = LocalOpenIdProvider,
    state: IExternalAuthorizationState = LocalAuthorizationRequestState,
    code: str | None = fastapi.Query(
        default=None,
        title='Authorization code',
        description=(
            'The authorization code generated by the authorization server. '
            'Required if using the Authorization Code Flow and the response '
            'mode is `query`, otherwise this parameter is ignored.'
        )
    ),
    iss: str | None = fastapi.Query(
        default=None,
        title='Issuer',
        description=(
            'Identifies the authorization server that redirected to '
            'this endpoint, as defined in RFC 9207. If the client supports '
            'OAuth 2.0 Authorization Server Issuer Identification, then '
            'this parameter is **required**, if the response mode is not '
            'JWT Secured Authorization Response Mode (JARM) per RFC 9101.\n\n'
            'The `iss` parameter value is the issuer identifier of '
            'the authorization server that created the authorization '
            'response, as defined in RFC 8414.  Its value **must** '
            'be a URL that uses the `https` scheme without any '
            'query or fragment components.'
        )
    ),
    error: str | None = fastapi.Query(
        default=None,
        title="Error code",
        description=(
            "The error code returned by the authorization server if "
            "the user cancelled the request, refused consent, or "
            "failed to authenticate."
        )
    )
) -> TokenResponse:
    if error:
        return None # type: ignore
    if code is None:
        raise FatalAuthorizationException(
            "The 'code' parameter is missing from the downstream "
            "response."
        )
    return await state.obtain(
        provider=provider,
        code=code,
        url=f'{request.url.scheme}://{request.url.netloc}{request.url.path}',
        iss=iss
    )

DownstreamTokenResponse: TokenResponse = fastapi.Depends(get)


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import copy
from inspect import Parameter
from string import Formatter
from typing import cast
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import TypeVar

import fastapi
import pydantic

from cbra.types import IEndpoint
from cbra.types import MutableSignature
from ..requesthandler import RequestHandler
from .iresource import IResource
from .resourceidentifier import ResourceIdentifier


T = TypeVar('T', bound='ResourceAction')


class ResourceAction(RequestHandler[IResource]): # type: ignore
    __module__: str = 'cbra.core'
    action: str
    name_template: str
    response_description: str = 'Successful response'
    status_code: int = 200

    @classmethod
    def fromfunc(
        cls: type[T],
        name: str,
        func: Any
    ) -> T:
        raise NotImplementedError

    @property
    def article(self) -> str:
        article = 'a'
        if self.name[0] in {'a', 'e', 'i', 'o', 'u'}:
            article = 'an'
        return article

    @property
    def name(self) -> str:
        return self.endpoint.model.__name__

    @property
    def response_model(self) -> type[pydantic.BaseModel]:
        return self.endpoint.model.__response_model__

    @property
    def path_parameter_name(self) -> str:
        return f'{self.endpoint.resource_name.lower()}_id'

    @property
    def summary(self) -> str:
        # TODO
        return self.name_template.format(
            article=self.article,
            name=self.endpoint.verbose_name,
            pluralname=self.endpoint.verbose_name_plural
        )

    def add_to_class(self, cls: type[IResource]) -> None: # type: ignore
        # Construct a resource identifier from the unprefixed path.
        formatter = Formatter()
        params: list[str] = [
            name for _, name, *_ in formatter.parse(self.get_url_pattern(None, cls))
            if name
        ]
        if params:
            ParentModel = cls.model.__key_model__
            parent_fields = list(ParentModel.__fields__.values())
            fields = {}
            annotations = {}
            for i, name in enumerate(params):
                field = copy.deepcopy(parent_fields[i])
                info = field.field_info
                info.update_from_config({'alias': name})
                annotations[name] = ParentModel.__annotations__[field.name]
                fields[name] = info
            Model = type(ParentModel.__name__, (ResourceIdentifier,), {
                '__annotations__': annotations,
                **fields
            })
            cls.resource_id = Model.depends()
            cls.__annotations__['resource_id'] = Model

        return super().add_to_class(cast(type[IEndpoint], cls))

    def add_to_router( # type: ignore
        self,
        cls: type[IResource],
        router: fastapi.APIRouter, **kwargs: Any
    ) -> None:
        kwargs.setdefault('status_code', self.status_code)
        kwargs.setdefault('summary', self.summary)
        tags: list[str] = kwargs.setdefault('tags', [])
        tags.append(self.name)
        kwargs.update({
            'path': self.get_url_pattern(kwargs.get('path')),
            'name': f'{cls.path_name}.{self.action}',
            'response_description': self.response_description.format(
                name=self.name,
                pluralname=self.endpoint.verbose_name_plural,
                article=self.article
            )
        })
        kwargs['responses'] = self.get_openapi_responses(cls, kwargs.get('responses') or {})
        return super().add_to_router(cls, router, **kwargs) # type: ignore

    def get_openapi_responses(
        self,
        cls: type[IResource],
        responses: dict[int | str, Any]
    ) -> dict[int | str, Any]:
        if getattr(cls, 'versioned', None):
            responses.setdefault(412, {
                'description': f"The **{cls.model.__name__}** was changed in between requests",
                'headers': {
                    'ETag': {
                        'schema': {'type': 'string'},
                        'description': (
                            "The most recent version of the resource. Use this (opaque) value "
                            "as the `If-Match` header on subsequent `POST`, `PUT` or `PATCH` "
                            "requests."
                        )
                    }
                }
            })
        return {
            **responses,
            **getattr(self._func, 'responses', {})
        }

    def get_url_pattern(self, prefix: str | None, endpoint: type[IResource] | None = None) -> str:
        endpoint = endpoint or self.endpoint
        path: str | None = prefix
        if path is None:
            path = f'/{endpoint.path_name}'
        else:
            path = f"{str.rstrip(path, '/')}/{endpoint.path_name}"
        assert str.startswith(path, '/')
        return path

    def get_write_model(self) -> type[pydantic.BaseModel]:
        raise NotImplementedError

    def is_detail(self) -> bool:
        raise NotImplementedError

    def needs_resource(self) -> bool:
        return False

    def parse_resource(self, resource: pydantic.BaseModel) -> pydantic.BaseModel:
        raise NotImplementedError

    def preprocess_parameter(self, p: Parameter) -> Parameter | None:
        """Hook to modify a parameter just before it is added to the
        new signature.
        """

    def preprocess_signature(
        self,
        parameters: dict[str, Parameter],
        return_annotation: Any
    ) -> tuple[list[Parameter], Any]:
        if self.can_write() and self.needs_resource():
            self._handler_params.add('resource')
            parameters['resource'] = Parameter(
                kind=Parameter.POSITIONAL_ONLY,
                name='resource',
                annotation=self.get_write_model(),
            )
        return super().preprocess_signature(parameters, return_annotation)

    def validate_handler(
        self,
        func: Callable[..., Awaitable[Any] | Any],
        signature: MutableSignature
    ) -> tuple[Callable[..., Awaitable[Any] | Any], MutableSignature]:
        if not signature.has_param('resource') and self.can_write() and self.needs_resource():
            raise TypeError(
                f"{self._endpoint_name}.{self.action} must accept "
                "the 'resource' positional argument."
            )
        return super().validate_handler(func, signature)

    async def preprocess_value(self, name: str, value: Any) -> Any:
        if name == 'resource' and self.needs_resource():
            value = self.parse_resource(value)
        return await super().preprocess_value(name, value)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic


class WebhookResponse(pydantic.BaseModel):
    accepted: bool = pydantic.Field(
        default=...,
        title="Accepted?",
        description='Indicates if the server accepted the webhook.'
    )
    success: bool = pydantic.Field(
        default=...,
        title="Succes?",
        description='Indicates if the server succesfully processed the message.'
    )

    reason: str | None = pydantic.Field(
        default=None,
        title="Reason",
        description='Describes why the webhook message was not accepted.'
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from cbra.types import Request


def get(
    request: Request,
    redirect_uri: str = fastapi.Cookie(
        default=...,
        title="Redirect URI",
        alias='bff.redirect_uri',
        description=(
            "The URI to redirect the user-agent to after completing the "
            "login."
        )
    )
) -> str:
    if str.startswith(redirect_uri, '/'):
        redirect_uri = f'{request.url.scheme}://{request.url.netloc}{redirect_uri}'
    return redirect_uri


FrontendRedirectURI: str = fastapi.Depends(get)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

from headless.ext.oauth2.models import TokenResponse

from .authorizationrequestidentifier import AuthorizationRequestIdentifier
from .clientidentifier import ClientIdentifier
from .objectidentifier import ObjectIdentifier
from .oidcprovider import OIDCProvider


class IExternalAuthorizationState(Protocol):

    @property
    def id(self) -> ObjectIdentifier['IExternalAuthorizationState']: ...

    @property
    def client_id(self) -> ClientIdentifier: ...

    @property
    def kind(self) -> str: ...

    @property
    def request_id(self) -> AuthorizationRequestIdentifier: ...

    async def obtain(self, provider: OIDCProvider, code: str, url: str, iss: str | None = None) -> TokenResponse: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets

import fastapi
from headless.ext.oauth2 import ClientAssertionType
from ckms.core.models import JOSEObject
from ckms.core.models import JSONWebSignature

import cbra.core as cbra
from cbra.types import Request
from ..authorizationserverstorage import AuthorizationServerStorage
from ..models import Client
from ..types import ClientAuthenticationMethod
from ..types import FatalClientException
from ..types import InvalidClient
from ..types import InvalidRequest


__all__: list[str] = [
    'RequestingClient'
]


MAX_CLIENT_ASSERTION_AGE: int = 60


async def get_from_client_assertion(
    request: Request,
    storage: AuthorizationServerStorage,
    assertion_type: ClientAssertionType | None,
    assertion: str,
) -> Client:
    if not assertion_type:
        raise InvalidRequest("The 'assertion_type' parameter is required.")
    if assertion_type != ClientAssertionType.jwt_bearer:
        raise InvalidRequest(f"Unsupported assertion type: {assertion_type[:64]}")
    try:
        jws = JOSEObject.parse(assertion)
    except Exception:
        raise InvalidClient("Unable to decode the JWT client assertion.")
    if not isinstance(jws, JSONWebSignature):
        raise InvalidClient(
            "The JOSE object in the client assertion must be of type "
            "JSON Web Signature (JWS)."
        )
    if str.lower(jws.typ or '') != 'jwt':
        raise InvalidClient(
            message=(
                "A JSON Web Token (JWT) enclosed in the assertion must be of "
                "type 'JWT'."
            )
        )
    try:
        jwt = jws.claims(accept={'jwt'})
    except Exception:
        raise InvalidClient("Unable to deserialize the JWT as JSON.")
    try:
        jwt.verify(
            audience={str(request.url)},
            max_age=MAX_CLIENT_ASSERTION_AGE
        )
    except Exception:
        raise InvalidClient(
            "The JSON Web Token (JWT) had an invalid audience, was "
            "expired or not effective, or otherwise did not satisfy "
            "the requirements specified in RFC 7523."
        )
    if jwt.iss != jwt.sub:
        raise InvalidClient(
            "The JSON Web Token (JWT) must be issued by the "
            "client."
        )

    client = await storage.get(Client, jwt.iss)
    if client is None:
        raise InvalidClient('The client does not exist.')
    if client.auth_method != ClientAuthenticationMethod.private_key:
        raise InvalidClient(
            "The client does not allow the provided authentication mode."
        )
    if client.jwks is None:
        raise InvalidClient(
            "No known keys could verify the signature of the JWS."
        )
    if not await jws.verify(client.jwks):
        raise InvalidClient(
            "No signatures in the JSON Web Signature (JWS) could "
            "be verified using the known keys of the client."
        )
    return client


async def get(
    request: Request,
    storage: AuthorizationServerStorage = cbra.instance('_AuthorizationServerStorage'),
    client_id: str | None = fastapi.Form(
        default=None,
        title="Client ID",
        description=(
            "**Required**, if the client is not authenticating with the "
            "authorization server or the chosen method of authentication is "
            "`client_secret_post`, `client_secret_jwt`, or `private_key_jwt`, "
            "otherwise this parameter is ignored."
        )
    ),
    client_secret: str | None = fastapi.Form(
        default=None,
        title="Client secret",
        description=(
            "The client-side secret. **Required** if the client is confidential and "
            "authenticates using `client_secret_post`."
        )
    ),
    assertion: str | None = fastapi.Form(
        default=None,
        title="Client assertion",
        alias='client_assertion',
        description=(
            "A digitally signed artifact that was created using a pre-registered "
            "asymmetric keypair. The `client_assertion_type` indicates the format "
            "and protocol used to create the assertion."
        )
    ),
    assertion_type: ClientAssertionType | None = fastapi.Form(
        default=None,
        title="Client assertion type",
        alias='client_assertion_type',
        description="The assertion type used by the client."
    )
) -> Client:
    if assertion is not None:
        return await get_from_client_assertion(request, storage, assertion_type, assertion)
    if client_id is None:
        raise InvalidClient
    client = await storage.get(Client, client_id)
    if client is None:
        raise FatalClientException(
            code='invalid_client',
            message='The client does not exist.'
        )
    if client.auth_method != ClientAuthenticationMethod.post:
        raise NotImplementedError
    # TODO: This is highly specific for client_secret_post
    if client.is_confidential() and not client_secret:
        raise FatalClientException(
            code='invalid_client',
            message="The 'client_secret' parameter is mandatory for this client."
        )
    secret = await client.get_client_secret()
    if not secrets.compare_digest(secret or '', client_secret or ''):
        raise FatalClientException(
            code='invalid_client',
            message="The provided client secret is not valid."
        )

    return client


RequestingClient: Client = fastapi.Depends(get)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any
from typing import Awaitable

import fastapi

from cbra.core.conf import settings
from cbra.types import Forbidden
from cbra.types import IAuthorizationContextFactory
from cbra.types import IDependant
from cbra.types import IRequestPrincipal
from cbra.types import ISubjectResolver
from cbra.types import UnauthenticatedAuthorizationContext
from ..ioc import instance
from ..utils import ensure_awaited
from .authenticatedcontext import AuthenticatedContext
from .authenticationservice import AuthenticationService
from .subject import Subject
from .subjectresolver import SubjectResolver


class AuthorizationContextFactory(IAuthorizationContextFactory, IDependant):
    """The default implementation of :class:`~cbra.types.IAuthorizationContextFactory`"""
    __module__: str = 'cbra.core.iam'
    authentication: AuthenticationService
    logger: logging.Logger = logging.getLogger('uvicorn')
    resolver: ISubjectResolver

    def __init__(
        self,
        resolver: ISubjectResolver = instance(
            name='SubjectResolver',
            missing=SubjectResolver
        ),
        authentication: AuthenticationService = instance(
            name='AuthenticationService',
            missing=AuthenticationService
        )
    ):
        self.authentication = authentication
        self.resolver = resolver

    def allowed_audience(self, request: fastapi.Request) -> set[str]:
        url = request.url
        allow = {
            f'{url.scheme}://{url.netloc}',
            str(url)
        }

        # If the ASGI_ROOT_PATH setting is defined, then we have a problem
        # with Eventarc and Cloud Scheduler, which use the URL without
        # the prefix. For now the solution is to also allow a URL with
        # the ASGI_ROOT_PATH stripped from the path, but its quite a
        # hack and must be fixed (TODO). Also we must validate that
        # the path starts with a slash somewhere else.
        if settings.ASGI_ROOT_PATH:
            if not str.startswith(settings.ASGI_ROOT_PATH, '/'):
                raise ValueError('ASGI_ROOT_PATH must begin with a slash.')
            if str.endswith(settings.ASGI_ROOT_PATH, '/'):
                raise ValueError('ASGI_ROOT_PATH must not end with a slash.')
            path = str.lstrip(url.path, settings.ASGI_ROOT_PATH)
            if not str.startswith(path, '/'):
                path = f'/{path}'
            allow.add(f'{url.scheme}://{url.netloc}{path}')
        return allow

    async def authenticate(
        self,
        request: fastapi.Request,
        principal: IRequestPrincipal,
        providers: set[str] | None = None,
        subjects: set[str] | Awaitable[set[str]] | None  = None,
        claims: dict[str, Any] | None = None
    ):
        remote_host = request.client.host if request.client else None
        subject = await principal.resolve(self.resolver.resolve)
        await subject.authenticate(
            self.authentication,
            providers=providers
        )
        if principal.has_audience():
            self.validate_audience(principal, self.allowed_audience(request))

        unauthenticated= UnauthenticatedAuthorizationContext(remote_host=remote_host)
        if not subject.is_authenticated():
            return unauthenticated

        # Determine if the subject is in the allowed subjects list.
        subjects = await ensure_awaited(subjects) or set()
        if subjects and subject.sub not in subjects:
            self.logger.critical(
                'Subject %s is not in the allowed subjects list',
                subject.sub
            )
            return unauthenticated

        return AuthenticatedContext(
            remote_host=remote_host,
            subject=subject
        )

    def _is_authenticated(
        self,
        request: fastapi.Request,
        subject: Subject,
        principal: IRequestPrincipal
    ) -> bool:
        return all([
            subject.is_authenticated(),
            self.is_authenticated(request, subject, principal)
        ])

    def is_authenticated(
        self,
        request: fastapi.Request,
        subject: Subject,
        principal: IRequestPrincipal
    ) -> bool:
        """Hook to determine if the request is succesfully authenticated.
        The default implementation always returns ``True``, but may be
        overriden to implement additional authenticated logic.
        """
        return True

    def validate_audience(self, principal: IRequestPrincipal, allow: set[str]):
        if not principal.validate_audience(allow):
            self.logger.critical(
                'Received a principal/credential combination with an '
                'audience that is not accepted by the server (allowed'
                ': %s, actual: %s)',
                str.join(', ', allow),
                str.join(', ', principal.get_audience()),
            )
            raise Forbidden


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from typing import Any
from typing import TypeVar

import pydantic
from google.cloud.datastore import Client as GoogleClient

import cbra.core as cbra
from cbra.core.iam.models import Principal
from cbra.core.iam.models import Subject
from cbra.ext.google import DatastoreSubjectRepository
from cbra.ext.oauth2 import BaseStorage
from cbra.ext.oauth2.models import AuthorizationRequest
from cbra.ext.oauth2.models import AuthorizationRequestParameters
from cbra.ext.oauth2.models import AuthorizationState
from cbra.ext.oauth2.models import Client
from cbra.ext.oauth2.models import ExternalAuthorizationState
from cbra.ext.oauth2.models import RefreshToken
from cbra.ext.oauth2.models import ResourceOwner
from cbra.ext.oauth2.types import AuthorizationCode
from cbra.ext.oauth2.types import AuthorizationRequestIdentifier
from cbra.ext.oauth2.types import ClientIdentifier
from cbra.ext.oauth2.types import ExtAuthorizationRequestState
from cbra.ext.oauth2.types import IssuedAccessToken
from cbra.ext.oauth2.types import IssuedAccessTokenIdentifier
from cbra.ext.oauth2.types import ObjectIdentifier
from cbra.ext.oauth2.types import RefreshTokenIdentifier
from cbra.ext.oauth2.types import ResourceOwnerIdentifier
from cbra.ext.oauth2.types import PairwiseIdentifier
from cbra.ext.oauth2.types import PrincipalIdentifier
from cbra.ext.security import ApplicationKeychain
from .basemodelstorage import BaseModelStorage


T = TypeVar('T', bound=pydantic.BaseModel)


class Storage(BaseStorage, BaseModelStorage):
    __module__: str = 'cbra.ext.google.impl.oauth2'
    keychain: ApplicationKeychain
    subjects: DatastoreSubjectRepository

    def __init__(
        self,
        client: GoogleClient | Any = cbra.inject('GoogleDatastoreClient'),
        keychain: ApplicationKeychain = ApplicationKeychain.depends()
    ):
        if not isinstance(client, GoogleClient):
            raise TypeError(f"Invalid client: {repr(client)}")
        self.client = client
        self.keychain = keychain
        self.subjects = DatastoreSubjectRepository(client)

    @functools.singledispatchmethod # type: ignore
    async def destroy(self, obj: Any) -> None:
        raise TypeError(type(obj).__name__)

    @destroy.register
    async def destroy_authorization_request(
        self,
        obj: AuthorizationRequest
    ) -> None:
        await self.delete(self.key('AuthorizationRequest', obj.id))

    @destroy.register
    async def destroy_refresh_token(
        self,
        obj: RefreshToken
    ):
        await self.delete(self.key('RefreshToken', str(obj.token)))

    @functools.singledispatchmethod # type: ignore
    async def fetch(self, key: ObjectIdentifier[T]) -> Any:
        raise TypeError(type(key).__name__)

    @fetch.register
    async def fetch_authorization_request(
        self,
        oid: AuthorizationRequestIdentifier
    ) -> AuthorizationRequest | None:
        key = self.key('AuthorizationRequest', str(oid))
        return self.entity_to_model(
            AuthorizationRequest,
            await self.get_entity_by_key(key),
            id=key.name
        )

    @fetch.register
    async def fetch_authorization_request_by_code(
        self,
        oid: AuthorizationCode
    ) -> AuthorizationRequest | None:
        q = self.query(kind='AuthorizationRequest')
        q.add_filter('code.value', '=', str(oid))
        entity = await self.first(q)
        if entity:
            return self.entity_to_model(
                AuthorizationRequest,
                entity,
                id=entity.key.name
            )

    @fetch.register
    async def fetch_client(
        self,
        oid: ClientIdentifier
    ) -> Client | None:
        return None

    @fetch.register
    async def fetch_external_state(
        self,
        oid: ExtAuthorizationRequestState
    ) -> ExternalAuthorizationState | None:
        key = self.key('ExternalAuthorizationState', str(oid))
        return self.entity_to_model(
            ExternalAuthorizationState,
            await self.get_entity_by_key(key),
            id=key.name
        )

    @fetch.register
    async def fetch_issued_access_token(
        self,
        oid: IssuedAccessTokenIdentifier
    ) -> IssuedAccessToken | None:
        key = self.key('IssuedAccessToken', str(oid))
        return self.entity_to_model(
            IssuedAccessToken,
            await self.get_entity_by_key(key),
            token_hash=key.name
        )
    
    @fetch.register
    async def fetch_principal(
        self,
        oid: PrincipalIdentifier
    ) -> None | Principal:
        key = self.key('Principal', str(oid))
        return self.entity_to_model(
            Principal,
            await self.get_entity_by_key(key),
            key=str(oid)
        )

    @fetch.register
    async def fetch_refresh_token(
        self,
        oid: RefreshTokenIdentifier
    ) -> RefreshToken | None:
        key = self.key('RefreshToken', str(oid))
        return self.entity_to_model(
            RefreshToken,
            await self.get_entity_by_key(key),
            token=key.name
        )

    async def get_authorization_request_by_code(
        self,
        oid: AuthorizationCode
    ) -> AuthorizationRequestParameters | None:
        q = self.query(kind='AuthorizationRequest')
        q.add_filter('code.value', '=', str(oid))
        entity = await self.first(q)
        if entity:
            return self.entity_to_model(
                AuthorizationRequestParameters,
                entity,
                id=entity.key.name
            )

    async def get_authorization_request_by_id(
        self,
        oid: AuthorizationRequestIdentifier
    ) -> AuthorizationRequestParameters | None:
        key = self.key('AuthorizationRequest', str(oid))
        return self.entity_to_model(
            AuthorizationRequestParameters,
            await self.get_entity_by_key(key),
            id=key.name
        )

    async def get_client(self, client_id: str) -> Client | None:
        return None

    async def get_resource_owner(
        self,
        oid: ResourceOwnerIdentifier
    ) -> ResourceOwner | None:
        key = self.key('ResourceOwner', f'clients/{oid.client_id}/subjects/{oid.sub}')
        return self.entity_to_model(
            ResourceOwner,
            await self.get_entity_by_key(key),
        )

    async def persist_authorization_request(self, obj: AuthorizationRequestParameters) -> None:
        entity = self.model_to_entity(self.key('AuthorizationRequest', obj.id), obj)
        await self.put(entity)

    async def persist_issued_access_token(
        self,
        obj: IssuedAccessToken
    ) -> None:
        entity = self.model_to_entity(
            self.key(type(obj).__name__, obj.token_hash),
            obj,
            exclude={'token_hash'}
        )
        await self.put(entity)

    async def persist_state(self, obj: AuthorizationState | ExternalAuthorizationState) -> None:
        if isinstance(obj, AuthorizationState):
            raise NotImplementedError
        entity = self.model_to_entity(self.key(type(obj).__name__, obj.id), obj.__root__)
        await self.put(entity)

    async def persist_ppid(self, obj: PairwiseIdentifier) -> None:
        key = self.key(
            'PairwiseIdentifier',
            await self.allocate('PairwiseIdentifier')
        )
        entity = self.entity_factory(key)
        entity.update({'sector': obj.sector, 'sub': obj.sub})
        await self.put(entity)
        assert entity.key.id is not None
        obj.value = entity.key.id

    async def persist_refresh_token(self, obj: RefreshToken) -> None:
        # TODO: Since we havent implement the AuthorizedGrant yet,
        # simply allocate an id.
        if obj.grant_id == 0:
            obj.grant_id = await self.allocate('AuthorizedGrant')
        key = self.key('RefreshToken', str(obj.token))
        await self.put(self.model_to_entity(key, obj, exclude={'token'}))

    async def persist_resource_owner(self, obj: ResourceOwner) -> None:
        key = self.key('ResourceOwner', f'clients/{obj.client_id}/subjects/{obj.sub}')
        await self.put(self.model_to_entity(key, obj))

    async def persist_subject(self, obj: Subject) -> None:
        if not obj.is_encypted():
            await obj.encrypt(self.keychain)
        await self.subjects.persist(obj)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse

import fastapi

from cbra.types import Abortable


class FrontendException(Abortable):
    __module__: str = 'cbra.ext.oauth2.types'
    return_url: str
    params: dict[str, str]

    def __init__(self, return_url: str, **params: str) -> None:
        self.params = params

        p: list[str | None] = list([
            str(x) if x else None
            for x in urllib.parse.urlparse(return_url)
        ])
        q: dict[str, str] = dict(urllib.parse.parse_qsl(p[4]))
        q.update(self.params)
        p = list(p)
        p[4] = urllib.parse.urlencode(q) # type: ignore
        self.return_url = urllib.parse.urlunparse(p)

    async def as_response(self) -> fastapi.Response:
        return fastapi.responses.RedirectResponse(
            status_code=303,
            url=self.return_url
        )
    


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from cbra.types import Abortable
from cbra.types import IDependant
from ..types import RedirectURI


DEFAULT_FRONTEND_ERROR_PAGE: str = '/error'


__all__: list[str] = [
    'DownstreamError'
]


class Error(IDependant, Abortable):
    error: str
    request: fastapi.Request
    return_url: RedirectURI | None = None

    def __init__(
        self,
        request: fastapi.Request,
        error: str,
        return_url: str | RedirectURI | None
    ):
        if return_url is not None and str.startswith(return_url, '/'):
            return_url = RedirectURI(
                f'{request.url.scheme}://{request.url.netloc}{return_url}'
            )
        if isinstance(return_url, str):
            return_url = RedirectURI(return_url)
        self.error = error
        self.return_url = return_url

    @classmethod
    def __inject__(cls): # type: ignore
        return cls.fromrequest
    
    @classmethod
    def fromrequest(
        cls,
        request: fastapi.Request,
        error: str | None = fastapi.Query(
            default=None,
            title="Error code",
            description=(
                "The error code returned by the authorization server if "
                "the user cancelled the request, refused consent, or "
                "failed to authenticate."
            )
        ),
        return_url: str | None = fastapi.Cookie(
            default=None,
            alias='oauth2.return_url',
            title="Return URL",
            description=(
                "A return URL where the server can redirect the user agent in "
                "case of an error."
            )
        )
    ):
        return cls(request, error, return_url=return_url) if error else None

    async def as_response(self) -> fastapi.Response:
        assert self.return_url is not None
        return fastapi.responses.RedirectResponse(
            status_code=303,
            url=self.return_url.redirect(allow_params=True, error=self.error)
        )


DownstreamError: Error | None = Error.depends()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from .endpoints import AuthorizationServerEndpoint


class CurrentSubjectEndpoint(AuthorizationServerEndpoint):
    name: str = 'oauth2.me'
    path: str = '/me'
    summary: str = 'Current User Endpoint'

    async def get(self) -> dict[str, Any]:
        await self.session
        claims: dict[str, Any] = {
            'sub': 'allUsers',
            'mailboxes': []
        }
        if self.is_authenticated():
            claims['sub'] = 'self'
            subject = await self.get_subject()
            for principal in subject.get_principals():
                if principal.spec.kind != 'EmailAddress':
                    continue
                claims['mailboxes'].append({
                    'principal_id': principal.key,
                    'email': principal.spec.email
                })

        return claims


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse

import fastapi

from cbra.core.conf import settings
from cbra.types import Abortable


class UserError(Abortable):
    __module__: str = 'cbra.ext.oauth2.types'
    authorization_url: str
    error: str
    request_id: str

    def __init__(
        self,
        error: str,
        request_id: str,
        authorize: str,
    ):
        self.authorization_url = authorize
        self.error = error
        self.request_id = request_id

    async def as_response(self) -> fastapi.Response:
        p: dict[str, str] = {
            'error': self.error,
            'next': self.authorization_url,
            'request': self.request_id
        }
        q = urllib.parse.urlencode(p, quote_via=urllib.parse.quote)
        return fastapi.responses.RedirectResponse(
            status_code=303,
            url=f'{settings.OAUTH2_ERROR_URL}?{q}'
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from headless.ext.oauth2.types import ResponseType

from .accesstype import AccessType
from .authorizationcode import AuthorizationCode
from .authorizationlifecycle import AuthorizationLifecycle
from .authorizationrequestidentifier import AuthorizationRequestIdentifier
from .bearertoken import BearerToken
from .bearertokencredential import BearerTokenCredential
from .bearertokenexception import BearerTokenException
from .clientauthenticationmethod import ClientAuthenticationMethod
from .clientidentifier import ClientIdentifier
from .clientinfo import ClientInfo
from .compositeobjectidentifier import CompositeObjectIdentifier
from .extauthorizationrequeststate import ExtAuthorizationRequestState
from .frontendexception import FrontendException
from .frontendobjecttype import FrontendObjectType
from .grantedscope import GrantedScope
from .iaccesstoken import IAccessToken
from .iaccesstokenobtainer import IAccessTokenObtainer
from .iauthorizationserverstorage import IAuthorizationServerStorage
from .fatalauthorizationexception import FatalAuthorizationException
from .fatalclientexception import FatalClientException
from .iauthorizationrequest import IAuthorizationRequest
from .invalidgrant import InvalidGrant
from .invalidtarget import InvalidTarget
from .iclient import IClient
from .iexternalauthorizationstate import IExternalAuthorizationState
from .ifrontendstorage import IFrontendStorage
from .imanagedgrant import IManagedGrant
from .invalidclient import InvalidClient
from .invalidrequest import InvalidRequest
from .invalidresponsetype import InvalidResponseTypeRequested
from .invalidscope import InvalidScope
from .irefreshtoken import IRefreshToken
from .iresourceowner import IResourceOwner
from .iresourceserverstorage import IResourceServerStorage
from .issuedaccesstoken import IssuedAccessToken
from .issuedaccesstokenidentifier import IssuedAccessTokenIdentifier
from .itokenbuilder import ITokenBuilder
from .itokensigner import ITokenSigner
from .jarmauthorizeresponse import JARMAuthorizeResponse
from .queryauthorizeresponse import QueryAuthorizeResponse
from .loginresponse import LoginResponse
from .managedgrantidentifier import ManagedGrantIdentifier
from .missingresponsetype import MissingResponseType
from .objectidentifier import ObjectIdentifier
from .oidcclaimset import OIDCClaimSet
from .oidcprovider import OIDCProvider
from .oidctokensubjectidentifier import OIDCTokenSubjectIdentifier
from .pairwiseidentifier import PairwiseIdentifier
from .principalidentifier import PrincipalIdentifier
from .redirecturi import RedirectURI
from .redirectparameters import RedirectParameters
from .refreshtokenidentifier import RefreshTokenIdentifier
from .refreshtokenpolicytype import RefreshTokenPolicyType
from .refreshtokentype import RefreshTokenType
from .requestedscope import RequestedScope
from .resourceaccesstokenidentifier import ResourceAccessTokenIdentifier
from .resourceserveraccesstoken import ResourceServerAccessToken
from .resourceowneridentifier import ResourceOwnerIdentifier
from .responsemodenotsupported import ResponseModeNotSupported
from .responsevalidationfailure import ResponseValidationFailure
from .rfc9068accesstoken import RFC9068AccessToken
from .signableoidctoken import SignableOIDCToken
from .unsupportedauthorizationresponse import UnsupportedAuthorizationResponse
from .usererror import UserError


__all__: list[str] = [
    'AccessType',
    'AuthorizationCode',
    'AuthorizationLifecycle',
    'AuthorizationRequestIdentifier',
    'BearerToken',
    'BearerTokenCredential',
    'BearerTokenException',
    'ClientIdentifier',
    'ClientInfo',
    'ClientAuthenticationMethod',
    'CompositeObjectIdentifier',
    'ExtAuthorizationRequestState',
    'FatalAuthorizationException',
    'FatalClientException',
    'FrontendException',
    'FrontendObjectType',
    'GrantedScope',
    'IAccessToken',
    'IAccessTokenObtainer',
    'IAuthorizationRequest',
    'IAuthorizationServerStorage',
    'IClient',
    'IExternalAuthorizationState',
    'IFrontendStorage',
    'IManagedGrant',
    'InvalidClient',
    'InvalidGrant',
    'InvalidRequest',
    'InvalidScope',
    'InvalidTarget',
    'InvalidResponseTypeRequested',
    'IRefreshToken',
    'IResourceOwner',
    'IResourceServerStorage',
    'IssuedAccessToken',
    'IssuedAccessTokenIdentifier',
    'ITokenBuilder',
    'ITokenSigner',
    'JARMAuthorizeResponse',
    'LoginResponse',
    'ManagedGrantIdentifier',
    'MissingResponseType',
    'ObjectIdentifier',
    'OIDCClaimSet',
    'OIDCProvider',
    'OIDCTokenSubjectIdentifier',
    'PairwiseIdentifier',
    'PrincipalIdentifier',
    'QueryAuthorizeResponse',
    'RedirectURI',
    'RedirectParameters',
    'RefreshTokenIdentifier',
    'RefreshTokenPolicyType',
    'RefreshTokenType',
    'RequestedScope',
    'ResourceAccessTokenIdentifier',
    'ResourceOwnerIdentifier',
    'ResourceServerAccessToken',
    'ResponseModeNotSupported',
    'ResponseType',
    'ResponseValidationFailure',
    'RFC9068AccessToken',
    'SignableOIDCToken',
    'UnsupportedAuthorizationResponse',
    'UserError',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime

import pydantic


class PicklistShipment(pydantic.BaseModel):
    idshipment: int
    idpicklist: int
    idorder: int
    idreturn: int | None = None
    idshippingprovider: int
    idcompany_shippingprovider: int
    idcompany_shippingprovider_profile: int
    provider: str
    providername: str
    public_providername: str
    profile_name: str
    carrier_key: str
    labelurl_pdf: str | None = None
    labelurl_zpl: str | None = None
    trackingcode: str | None = None
    trackingurl: str | None = None
    tracktraceurl: str | None = None
    created_by_iduser: int
    cancelled: bool = False
    created: datetime.datetime
    updated: datetime.datetime


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import timedelta
from typing import Any

import fastapi
from cbra.core.conf import settings
from google.cloud import storage

from .applicationstorageclient import ApplicationStorageClient


__all__: list[str] = [
    'ApplicationStorageBucket'
]


class StorageBucket:

    def __init__(self, bucket: storage.Bucket) -> None:
        self.bucket = bucket

    async def blob(self, path: str) -> Any:
        obj = self.bucket.blob(path) # type: ignore
        obj.reload() # type: ignore
        return obj

    async def generate_signed_url(
        self,
        path: str,
        expires: timedelta | None = None,
        ttl: int | None = None,
        *args: Any,
        **kwargs: Any
    ) -> str:
        if not bool(ttl) ^ bool(expires):
            raise TypeError("Provider either expires or ttl")
        if ttl is not None:
            expires = timedelta(seconds=ttl)
        blob = self.bucket.blob(path) # type: ignore
        return blob.generate_signed_url( # type: ignore
            expiration=expires,
            *args,
            **kwargs
        )


async def get(client: storage.Client = ApplicationStorageClient) -> storage.Bucket:
    return StorageBucket(client.bucket(settings.APP_STORAGE_BUCKET)) # type: ignore


ApplicationStorageBucket: StorageBucket = fastapi.Depends(get)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect

import ckms.utils
import pydantic
from ckms.core import KeySpecification
from ckms.types import CipherText


class VersionedCipherText(pydantic.BaseModel):
    aad: str | None = None
    ct: str
    dsn: str
    enc: str | None = None
    iv: str | None = None
    tag: str | None = None

    @pydantic.validator('ct', 'iv', 'aad', 'tag', pre=True)
    def preprocess_bytes(cls, value: bytes | str | None) -> str | None:
        if isinstance(value, bytes):
            value = ckms.utils.b64encode_str(value)
        return value
    
    async def decrypt(self, key: KeySpecification) -> bytes:
        params: dict[str, bytes] = {}
        if self.aad: params['aad'] = ckms.utils.b64decode(self.aad)
        if self.iv: params['iv'] = ckms.utils.b64decode(self.iv)
        if self.tag: params['tag'] = ckms.utils.b64decode(self.tag)
        ct = CipherText(buf=ckms.utils.b64decode(self.ct), **params)
        result = key.decrypt(ct)
        if inspect.isawaitable(result):
            result = await result
        assert isinstance(result, bytes)
        return result
        


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets

import pytest
import pydantic

from cbra.types import BaseModel


# Set defaults on the test model so we can always instantiate
# them.
class ModelWithScalarIdentity(BaseModel):
    id: int = pydantic.Field(1, primary_key=True)
    non_key_field: str = pydantic.Field(
        default_factory=lambda: secrets.token_urlsafe(48)
    )


class ModelWithCompositeIdentity(BaseModel):
    id: int = pydantic.Field(1, primary_key=True)
    item_id: int = pydantic.Field(2, primary_key=True)
    non_key_field: str = pydantic.Field(
        default_factory=lambda: secrets.token_urlsafe(48)
    )


class ModelWithNestedIdentity(BaseModel):
    id: int = pydantic.Field(1, primary_key=True)
    child: ModelWithScalarIdentity = pydantic.Field(
        default_factory=ModelWithScalarIdentity,
        primary_key=True
    )


class ModelWithAutoAssignedIdentity(BaseModel):
    id: int = pydantic.Field(1, primary_key=True)
    auto: int | None = pydantic.Field(None, auto_assign=True)


@pytest.mark.parametrize("Model", [
    ModelWithScalarIdentity,
    ModelWithCompositeIdentity,
    ModelWithNestedIdentity,
    ModelWithAutoAssignedIdentity
])
def test_model_has_key(Model: type[BaseModel]):
    assert hasattr(Model, '__key__')


@pytest.mark.parametrize("Model", [
    ModelWithScalarIdentity,
    ModelWithCompositeIdentity,
    ModelWithNestedIdentity,
    ModelWithAutoAssignedIdentity
])
def test_eq(Model: type[BaseModel]):
    obj1 = Model()
    obj2 = Model.parse_obj({
        k: v for k, v in obj1.__key__
    })
    assert obj1 == obj2


@pytest.mark.parametrize("Model", [
    ModelWithScalarIdentity,
    ModelWithCompositeIdentity,
    ModelWithNestedIdentity,
    ModelWithAutoAssignedIdentity
])
def test_neq(Model: type[BaseModel]):
    obj1 = Model()
    obj2 = Model(id=secrets.choice(range(0, 1000)))
    assert obj1 != obj2, f'{obj1.__key__} should not equal {obj2.__key__}'


@pytest.mark.parametrize("Model", [
    ModelWithScalarIdentity,
    ModelWithCompositeIdentity,
    ModelWithNestedIdentity,
    ModelWithAutoAssignedIdentity
])
def test_contains(Model: type[BaseModel]):
    objects = [
        Model(),
        Model(id=secrets.choice(range(0, 1000))),
        Model(id=secrets.choice(range(0, 1000))),
        Model(id=secrets.choice(range(0, 1000))),
    ]
    obj = Model()
    assert obj in objects
    assert len(list(filter(lambda x: x==obj, objects))) == 1


@pytest.mark.parametrize("Model", [
    ModelWithScalarIdentity,
    ModelWithCompositeIdentity,
    ModelWithNestedIdentity,
    ModelWithAutoAssignedIdentity
])
def test_not_contains(Model: type[BaseModel]):
    objects = [
        Model(id=secrets.choice(range(0, 1000))),
        Model(id=secrets.choice(range(0, 1000))),
        Model(id=secrets.choice(range(0, 1000))),
    ]
    obj = Model()
    assert obj not in objects
    assert len(list(filter(lambda x: x==obj, objects))) == 0


@pytest.mark.parametrize("Model", [
    ModelWithScalarIdentity,
    ModelWithCompositeIdentity,
    ModelWithNestedIdentity,
    ModelWithAutoAssignedIdentity
])
def test_list_remove(Model: type[BaseModel]):
    objects = [
        Model(),
        Model(),
        Model(id=secrets.choice(range(0, 1000))),
        Model(id=secrets.choice(range(0, 1000))),
        Model(id=secrets.choice(range(0, 1000))),
    ]
    objects.remove(Model())
    assert len(objects) == 4
    objects.remove(Model())
    assert len(objects) == 3


@pytest.mark.parametrize("Model", [
    ModelWithScalarIdentity,
    ModelWithCompositeIdentity,
    ModelWithNestedIdentity,
    ModelWithAutoAssignedIdentity
])
def test_set_types(Model: type[BaseModel]):
    objects = {
        Model(),
        Model(),
        Model(),
    }
    assert len(objects) == 1


def test_get_auto_assigned_field():
    instance = ModelWithAutoAssignedIdentity()
    assert hasattr(instance, '__surrogate__')
    assert instance.__surrogate__ is None
    instance.auto = 2
    assert instance.__surrogate__ == 2